/*
Module : MSMQWrappers.h
Purpose: Defines the interface for a set of C++ wrapper classes which encapsulate Microsoft Message Queuing (MSMQ).
History: PJN / 05-05-2014 1. Initial public release
         PJN / 22-11-2015 1. Updated copyright details.
                          2. Updated the code to clean compile at warning level 4
                          3. Updated the code to clean compile on VC 2015
                          4. Added comprehensive SAL annotations to all the code
                          5. Replaced all occurrences of NULL with nullptr throughout the code
         PJN / 19-12-2017 1. Updated copyright details.
                          2. Updated the code to compile cleanly when _ATL_NO_AUTOMATIC_NAMESPACE is defined.
                          3. Verified there has been no further additions to MSMQ from the latest Windows 10 SDK
                          4. Fixed up a compilation warning in VC 2015 in the MSMQDemo.cpp module.
         PJN / 22-09-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean
                          that the code will now only compile on VC 2017 or later.
         PJN / 20-04-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019
         PJN / 15-09-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 18-12-2019 1. Updated initialization of various structs to use C++ 11 list initialization
                          2. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 02-02-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 31-10-2021 1. Updated copyright details.
                          2. Fixed incorrect fallthrough in switch statement in CPropVariantAutoFree::FreeMemory.
         PJN / 19-02-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations
         PJN / 18-05-2023 1. Updated copyright details.
                          2. Updated modules to indicate that it needs to be compiled using /std:c++17. Thanks
                          to Martin Richter for reporting this issue.

Copyright (c) 2014 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


////////////////////////// Macros / Defines ///////////////////////////////////

#pragma once

#if _MSVC_LANG < 201703
#error MSMQWrappers requires a minimum C++ language standard of /std:c++17
#endif //#if _MSVC_LANG < 201703

#ifndef __MSMQWRAPPERS_H__
#define __MSMQWRAPPERS_H__

#ifndef CMSMQWRAPPERS_EXT_CLASS
#define CMSMQWRAPPERS_EXT_CLASS
#endif

#pragma comment(lib, "MqRt.lib") //Automatically link in the MSMQ dll


////////////////////////// Includes ///////////////////////////////////////////

#include <string>

#ifndef __ATLDEF_H__
#pragma message("To avoid this message, please put atldef.h in your pre compiled header (normally stdafx.h)")
#include <atldef.h>
#endif //#ifndef __ATLDEF_H__

#include "PropVariantWrapper.h"

#ifndef __MQ_H__
#pragma message("To avoid this message, please put Mq.h in your pre compiled header (normally stdafx.h)")
#include <Mq.h>
#endif //#ifndef __MQ_H__


////////////////////////// Classes ////////////////////////////////////////////

namespace MSMQ
{

//Typedefs
using String = std::wstring;


//A version of the CComPropVariant which automatically calls MQFreeMemory in its destructor
class CPropVariantAutoFree : public ATL::CComPropVariant
{
public:
//Constructors / Destructors
  CPropVariantAutoFree() noexcept
  {
    //Auto set the variant type to VT_NULL
    vt = VT_EMPTY;
  }

  CPropVariantAutoFree(_In_ const CPropVariantAutoFree&) = delete;
  CPropVariantAutoFree(_In_ CPropVariantAutoFree&&) = delete;

  ~CPropVariantAutoFree() noexcept
  {
    FreeMemory(*this);
  }

  //Note we do not need to reimplement any of the CComPropVariant constructors because
  //we will always be using the default constructor to initialize the PROPVARIANT 
  //to VT_NULL. Otherwise you could just use CComPropVariant directly

//Methods
  CPropVariantAutoFree& operator=(_In_ const CPropVariantAutoFree&) = delete;
  CPropVariantAutoFree& operator=(_In_ CPropVariantAutoFree&&) = delete;

#pragma warning(suppress: 26487)
  static HRESULT FreeMemory(_In_ PROPVARIANT& propVariant) noexcept
  {
    //This handles all the documented MSMQ properties which can be freed via MQFreeMemory. These
    //values are documented on the Microsoft docs as "MSMQ Queue Properties", "MSMQ Message Properties",
    //"MSMQ Private Computer Properties", "MSMQ Queue Manager Properties" & "MSMQ Management Properties".
    //Please note that this is not a comprehensive implementation and does not handle all
    //variant types such as VT_LPSTR or VT_VECTOR | VT_UI2 etc, etc. The variant types handled is up 
    //to date as of Q1 2022, the Windows 10.0.22543 SDK and MSMQ 3.0 which is the latest revision at that time.

    //What will be the return value from this method (assume the best)
    HRESULT hr{S_OK};

    switch (propVariant.vt)
    {
      case VT_NULL:
      {
        break; //Nothing to do
      }
      case VT_EMPTY:
      {
        break; //Nothing to do
      }
      case VT_LPWSTR:
      {
        MQFreeMemory(propVariant.pwszVal);
        break;
      }
      case VT_CLSID:
      {
        MQFreeMemory(propVariant.puuid);
        break;
      }
      case VT_BLOB:
      {
        MQFreeMemory(propVariant.blob.pBlobData);
        break;
      }
      case VT_VECTOR | VT_CLSID:
      {
        MQFreeMemory(propVariant.cauuid.pElems);
        break;
      }
      case VT_VECTOR | VT_UI1:
      {
        MQFreeMemory(propVariant.caub.pElems);
        break;
      }
      case VT_VECTOR | VT_UI8:
      {
        MQFreeMemory(propVariant.cauh.pElems);
        break;
      }
      case VT_VECTOR | VT_I4:
      {
        MQFreeMemory(propVariant.cal.pElems);
        break;
      }
      case VT_VECTOR | VT_UI4:
      {
        MQFreeMemory(propVariant.caul.pElems);
        break;
      }
      case VT_VECTOR | VT_LPWSTR:
      {
        for (ULONG i=0; i<propVariant.calpwstr.cElems; i++)
        {
#pragma warning(suppress: 26477)
          ATLASSERT(propVariant.calpwstr.pElems != nullptr);
#pragma warning(suppress: 26481 26486)
          MQFreeMemory(propVariant.calpwstr.pElems[i]); //Free each element
        }
        MQFreeMemory(propVariant.calpwstr.pElems);
        break;
      }
      case VT_VECTOR | VT_VARIANT:
      {
        for (ULONG i=0; i<propVariant.capropvar.cElems; i++)
        {
#pragma warning(suppress: 26477)
          ATLASSERT(propVariant.capropvar.pElems != nullptr);
#pragma warning(suppress: 26481 26486)
          const HRESULT hr2{FreeMemory(propVariant.capropvar.pElems[i])}; //Free each element

          //Update this methods return value if the call to FreeMemory above failed
          if (FAILED(hr2))
            hr = hr2; //NOLINT(clang-analyzer-deadcode.DeadStores)
        }
        MQFreeMemory(propVariant.capropvar.pElems);
        break;
      }
      default:
      {
        //We have been given an unexpected vt value
#pragma warning(suppress: 26477)
        ATLASSERT(FALSE);
        hr = E_UNEXPECTED;
        break;
      }
    }

    //Reset the variant type now that we have freed up the memory
    propVariant.vt = VT_EMPTY;

    return hr;
  }
};


//Encapsulates a MSMQ cursor
class CCursor
{
public:
//Constructors / Destructors
  CCursor() noexcept : m_hCursor{nullptr}
  {
  }

  CCursor(_In_ const CCursor&) = delete;

  CCursor(_In_ CCursor&& cursor) noexcept : m_hCursor{nullptr}
  {
    Attach(cursor.Detach());
  }

  explicit CCursor(_In_ HANDLE hCursor) noexcept : m_hCursor{hCursor}
  {
  }

  ~CCursor() noexcept
  {
    if (m_hCursor != nullptr)
      Close();
  }

//Methods
  CCursor& operator=(_In_ const CCursor&) = delete;

  CCursor& operator=(_In_ CCursor&& cursor) noexcept
  {
    if (m_hCursor != nullptr)
      Close();
    Attach(cursor.Detach());
    return *this;
  }

  HRESULT Create(_In_ QUEUEHANDLE hQueue) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCursor == nullptr);

    return MQCreateCursor(hQueue, &m_hCursor);
  }

  HRESULT Close() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCursor != nullptr);

    const HRESULT hr{MQCloseCursor(m_hCursor)};
    m_hCursor = nullptr;
    return hr;
  }

  operator HANDLE() const noexcept
  {
    return m_hCursor;
  }

  void Attach(_In_ HANDLE hCursor) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hCursor == nullptr);

    m_hCursor = hCursor;
  }

  HANDLE Detach() noexcept
  {
    HANDLE hCursor{m_hCursor};
    m_hCursor = nullptr;
    return hCursor;
  }

protected:
//Member variables
  HANDLE m_hCursor;
};


//Encapsulates a MSMQ handle returned by MQGetSecurityContext(Ex)
class CSecurityContext
{
public:
//Constructors / Destructors
  CSecurityContext() noexcept : m_hSecurityContext{nullptr}
  {
  }

  CSecurityContext(_In_ const CSecurityContext&) = delete;

  CSecurityContext(_In_ CSecurityContext&& sc) noexcept : m_hSecurityContext{nullptr}
  {
    Attach(sc.Detach());
  }

  explicit CSecurityContext(_In_ HANDLE hSecurityContext) noexcept : m_hSecurityContext{hSecurityContext}
  {
  }

  ~CSecurityContext() noexcept
  {
    if (m_hSecurityContext != nullptr)
      Free();
  }

//Methods
  CSecurityContext& operator=(_In_ const CSecurityContext&) = delete;

  CSecurityContext& operator=(_In_ CSecurityContext&& sc) noexcept
  {
    if (m_hSecurityContext != nullptr)
      Free();
    Attach(sc.Detach());
    return *this;
  }

  HRESULT Get(_In_reads_bytes_opt_(dwCertBufferLength) LPVOID lpCertBuffer, _In_ DWORD dwCertBufferLength) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hSecurityContext == nullptr);

    return MQGetSecurityContextEx(lpCertBuffer, dwCertBufferLength, &m_hSecurityContext);
  }

  void Free() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hSecurityContext != nullptr);

    MQFreeSecurityContext(m_hSecurityContext);
    m_hSecurityContext = nullptr;
  }

  operator HANDLE() const noexcept
  {
    return m_hSecurityContext;
  }

  void Attach(_In_ HANDLE hSecurityContext) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hSecurityContext == nullptr);

    m_hSecurityContext = hSecurityContext;
  }

  HANDLE Detach() noexcept
  {
    HANDLE hSecurityContext{m_hSecurityContext};
    m_hSecurityContext = nullptr;
    return hSecurityContext;
  }

protected:
//Member variables
  HANDLE m_hSecurityContext;
};


//Encapsulates a MSMQ handle returned by MQLocateBegin
class CSearch
{
public:
//Constructors / Destructors
  CSearch() noexcept : m_hEnum{nullptr}
  {
  }

  CSearch(_In_ const CSearch&) = delete;

  CSearch(_In_ CSearch&& search) noexcept : m_hEnum{nullptr}
  {
    Attach(search.Detach());
  }

  explicit CSearch(_In_ HANDLE hEnum) noexcept : m_hEnum{hEnum}
  {
  }

  ~CSearch() noexcept
  {
    if (m_hEnum != nullptr)
      LocateEnd();
  }

//Methods
  CSearch& operator=(_In_ const CSearch&) = delete;

  CSearch& operator=(_In_ CSearch&& search) noexcept
  {
    if (m_hEnum != nullptr)
      LocateEnd();
    Attach(search.Detach());
    return *this;
  }

  HRESULT LocateBegin(_In_opt_ MQRESTRICTION* pRestriction, _In_ MQCOLUMNSET* pColumns, _In_ MQSORTSET* pSort) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hEnum == nullptr);

    return MQLocateBegin(nullptr, pRestriction, pColumns, pSort, &m_hEnum);
  }

#pragma warning(suppress: 26485)
  HRESULT LocateNext(_Inout_ DWORD* pcProps, _Out_ MQPROPVARIANT aPropVar[]) noexcept //NOLINT(modernize-avoid-c-arrays)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hEnum != nullptr);
  
    return MQLocateNext(m_hEnum, pcProps, aPropVar);
  }

  HRESULT LocateEnd() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hEnum != nullptr);

    return MQLocateEnd(m_hEnum);
  }

  operator HANDLE() const noexcept
  {
    return m_hEnum;
  }

  void Attach(_In_ HANDLE hEnum) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hEnum == nullptr);

    m_hEnum = hEnum;
  }

  HANDLE Detach() noexcept
  {
    HANDLE hEnum{m_hEnum};
    m_hEnum = nullptr;
    return hEnum;
  }

protected:
//Member variables
  HANDLE m_hEnum;
};


//Encapsulates a MSMQ "QUEUEHANDLE"
class CQueue
{
public:
//Constructors / Destructors
  CQueue() noexcept : m_hQueue{nullptr}
  {
  }

  CQueue(_In_ const CQueue&) = delete;

  CQueue(_In_ CQueue&& queue) noexcept : m_hQueue{nullptr}
  {
    Attach(queue.Detach());
  }
  
  explicit CQueue(_In_ QUEUEHANDLE hQueue) noexcept : m_hQueue{hQueue}
  {
  }

  ~CQueue() noexcept
  {
    if (m_hQueue != nullptr)
      Close();
  }

//Methods
  CQueue& operator=(_In_ const CQueue&) = delete;

  CQueue& operator=(_In_ CQueue&& queue) noexcept
  {
    if (m_hQueue != nullptr)
      Close();
    Attach(queue.Detach());
    return *this;
  }

  static HRESULT Create(_Inout_ MQQUEUEPROPS* pQueueProps, _Out_ String& sFormatName, _In_opt_ PSECURITY_DESCRIPTOR pSecurityDescriptor = nullptr)
  {
    DWORD dwFormatNameLength{_MAX_PATH};
    sFormatName.resize(dwFormatNameLength);
#pragma warning(suppress: 26446 26489)
    const HRESULT hr{MQCreateQueue(pSecurityDescriptor, pQueueProps, sFormatName.data(), &dwFormatNameLength)};
    if (SUCCEEDED(hr))
#pragma warning(suppress: 26489)
      sFormatName.resize(dwFormatNameLength - 1);
    else
#pragma warning(suppress: 26489)
      sFormatName.clear();
    return hr;
  }

  static HRESULT ADsPathToFormatName(_In_ LPCWSTR pwszADsPath, _Out_ String& sFormatName)
  {
    DWORD dwFormatNameLength{_MAX_PATH};
    sFormatName.resize(dwFormatNameLength);
#pragma warning(suppress: 26446 26489)
    const HRESULT hr{MQADsPathToFormatName(pwszADsPath, sFormatName.data(), &dwFormatNameLength)};
    if (SUCCEEDED(hr))
#pragma warning(suppress: 26489)
      sFormatName.resize(dwFormatNameLength - 1);
    else
#pragma warning(suppress: 26489)
      sFormatName.clear();
    return hr;
  }

  static HRESULT InstanceToFormatName(_In_ GUID& guid, _Out_ String& sFormatName)
  {
    DWORD dwFormatNameLength{_MAX_PATH};
    sFormatName.resize(dwFormatNameLength);
#pragma warning(suppress: 26446 26489)
    const HRESULT hr{MQInstanceToFormatName(&guid, sFormatName.data(), &dwFormatNameLength)};
    if (SUCCEEDED(hr))
#pragma warning(suppress: 26489)
      sFormatName.resize(dwFormatNameLength - 1);
    else
#pragma warning(suppress: 26489)
      sFormatName.clear();
    return hr;
  }

  static HRESULT PathNameToFormatName(_In_ LPCWSTR lpwcsPathName, _Out_ String& sFormatName)
  {
    DWORD dwFormatNameLength{_MAX_PATH};
    sFormatName.resize(dwFormatNameLength);
#pragma warning(suppress: 26446 26489)
    const HRESULT hr{MQPathNameToFormatName(lpwcsPathName, sFormatName.data(), &dwFormatNameLength)};
    if (SUCCEEDED(hr))
#pragma warning(suppress: 26489)
      sFormatName.resize(dwFormatNameLength - 1);
    else
#pragma warning(suppress: 26489)
      sFormatName.clear();
    return hr;
  }

  static HRESULT Delete(_In_ LPCWSTR pwcsFormatName) noexcept
  {
    return MQDeleteQueue(pwcsFormatName);
  }

  HRESULT Open(_In_ LPCWSTR lpwcsFormatName, _In_ DWORD dwAccess, _In_ DWORD dwShareMode = MQ_DENY_NONE) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue == nullptr);

    return MQOpenQueue(lpwcsFormatName, dwAccess, dwShareMode, &m_hQueue);
  }

  HRESULT Close() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);

    const HRESULT hr{MQCloseQueue(m_hQueue)};
    m_hQueue = nullptr;
    return hr;
  }

  HRESULT GetFormatName(_Out_ String& sFormatName)
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);

    DWORD dwFormatNameLength{_MAX_PATH};
    sFormatName.resize(dwFormatNameLength);
#pragma warning(suppress: 26446 26489)
    const HRESULT hr{MQHandleToFormatName(m_hQueue, sFormatName.data(), &dwFormatNameLength)};
    if (SUCCEEDED(hr))
#pragma warning(suppress: 26489)
      sFormatName.resize(dwFormatNameLength - 1);
    else
#pragma warning(suppress: 26489)
      sFormatName.clear();
    return hr;
  }

  HRESULT Receive(_In_ DWORD dwTimeout, _In_ DWORD dwAction, _Inout_opt_ MQMSGPROPS* pMessageProps, _Inout_opt_ LPOVERLAPPED lpOverlapped = nullptr, _In_opt_ PMQRECEIVECALLBACK fnReceiveCallback = nullptr, _In_opt_ HANDLE hCursor = nullptr, _In_opt_ ITransaction* pTransaction = nullptr) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);

    return MQReceiveMessage(m_hQueue, dwTimeout, dwAction, pMessageProps, lpOverlapped, fnReceiveCallback, hCursor, pTransaction);
  }

  HRESULT ReceiveByLookupId(_In_ ULONGLONG ullLookupId, _In_ DWORD dwLookupAction, _Inout_opt_ MQMSGPROPS* pMessageProps, _Inout_opt_ LPOVERLAPPED lpOverlapped = nullptr, _In_opt_ PMQRECEIVECALLBACK fnReceiveCallback = nullptr, _In_opt_ ITransaction* pTransaction = nullptr) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);

    return MQReceiveMessageByLookupId(m_hQueue, ullLookupId, dwLookupAction, pMessageProps, lpOverlapped, fnReceiveCallback, pTransaction);
  }

  HRESULT Move(_In_ CQueue& targetQueue, _In_ ULONGLONG lookupID, _In_opt_ ITransaction* pTransaction = nullptr) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);
#pragma warning(suppress: 26477)
    ATLASSERT(targetQueue.m_hQueue != nullptr);

    return MQMoveMessage(m_hQueue, targetQueue.m_hQueue,lookupID, pTransaction);
  }

  HRESULT Purge() noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);

    return MQPurgeQueue(m_hQueue);
  }

  HRESULT MarkRejected(_In_ ULONGLONG ullLookupId) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);

    return MQMarkMessageRejected(m_hQueue, ullLookupId);
  }

  HRESULT Send(_In_ MQMSGPROPS* pMessageProps, _In_opt_ ITransaction* pTransaction = nullptr) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue != nullptr);

    return MQSendMessage(m_hQueue, pMessageProps, pTransaction);
  }

  operator QUEUEHANDLE() const noexcept
  {
    return m_hQueue;
  }

  void Attach(_In_ QUEUEHANDLE hQueue) noexcept
  {
    //Validate our parameters
#pragma warning(suppress: 26477)
    ATLASSERT(m_hQueue == nullptr);

    m_hQueue = hQueue;
  }

  QUEUEHANDLE Detach() noexcept
  {
    HANDLE hQueue{m_hQueue};
    m_hQueue = nullptr;
    return hQueue;
  }

protected:
//Member variables
  QUEUEHANDLE m_hQueue;
};


}; //namespace MSMQ

#endif //#ifndef __MSMQWRAPPERS_H__
